import click
from simple_zarr_server.server import serve
import zarr
from uvicorn.config import (
    HTTP_PROTOCOLS,
    LIFESPAN,
    LOOP_SETUPS,
    WS_PROTOCOLS,
)

HTTP_CHOICES = click.Choice(HTTP_PROTOCOLS.keys())
WS_CHOICES = click.Choice(WS_PROTOCOLS.keys())
LIFESPAN_CHOICES = click.Choice(LIFESPAN.keys())
LOOP_CHOICES = click.Choice([key for key in LOOP_SETUPS.keys() if key != "none"])


@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.option(
    "--cors",
    type=str,
    default=None,
    help="Origin to allow CORS. Use wildcard '*' to allow all.",
)
@click.option(
    "--allow-write",
    "-w",
    help="Whether to allow PUT requests and enable write to underlying store.",
    is_flag=True,
)
# These options are mostly copied from uvicorn's docs
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this host.",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=8000,
    help="Bind socket to this port.",
    show_default=True,
)
@click.option(
    "--debug", is_flag=True, default=False, help="Enable debug mode.", hidden=True
)
@click.option("--reload", is_flag=True, default=False, help="Enable auto-reload.")
@click.option(
    "--loop",
    type=LOOP_CHOICES,
    default="auto",
    help="Event loop implementation.",
    show_default=True,
)
@click.option(
    "--http",
    type=HTTP_CHOICES,
    default="auto",
    help="HTTP protocol implementation.",
    show_default=True,
)
@click.option(
    "--ws",
    type=WS_CHOICES,
    default="auto",
    help="WebSocket protocol implementation.",
    show_default=True,
)
@click.option(
    "--use-colors/--no-use-colors",
    is_flag=True,
    default=None,
    help="Enable/Disable colorized logging.",
)
@click.option(
    "--proxy-headers/--no-proxy-headers",
    is_flag=True,
    default=True,
    help="Enable/Disable X-Forwarded-Proto, X-Forwarded-For, X-Forwarded-Port to populate remote address info.",
)
@click.option(
    "--forwarded-allow-ips",
    type=str,
    default=None,
    help="Comma seperated list of IPs to trust with proxy headers. Defaults to the $FORWARDED_ALLOW_IPS environment variable if available, or '127.0.0.1'.",
)
def main(
    path,
    cors,
    allow_write,
    host,
    port,
    loop,
    http,
    ws,
    debug,
    reload,
    proxy_headers,
    forwarded_allow_ips,
    use_colors,
):
    # Handles opening .n5, .zip, and .zarr extensions
    z = zarr.open(path, mode="a" if allow_write else "r")
    config = {
        "host": host,
        "port": port,
        "loop": loop,
        "http": http,
        "ws": ws,
        "debug": debug,
        "reload": reload,
        "proxy_headers": proxy_headers,
        "forwarded_allow_ips": forwarded_allow_ips,
        "use_colors": use_colors,
    }
    serve(z, allowed_origins=[cors], **config)



import pytest
import numpy as np
import zarr
from simple_zarr_server.server import create_zarr_server
from starlette.testclient import TestClient


class HTTPStore:
    def __init__(self, client):
        self.client = client

    def __getitem__(self, path):
        r = self.client.get(f"/{path}")
        if r.ok:
            return r.content
        else:
            raise KeyError

    def __contains__(self, path):
        r = self.client.head(f"/{path}")
        if r.ok:
            return True

    def __setitem__(self, path, cdata):
        r = self.client.put(f"/{path}", cdata)
        if not r.ok:
            raise ValueError


def test_numpy_read_only():
    # Create data
    original = np.random.rand(1024, 1024)
    z = zarr.array(original, read_only=True)

    # Initialize app
    app = create_zarr_server(z)

    # Open remote array and compare
    remote_store = HTTPStore(TestClient(app))
    arr = zarr.open_array(remote_store)
    np.testing.assert_allclose(arr[:], original)

    # Make sure can't write
    with pytest.raises(ValueError):
        arr[:50, :50] = 10


def test_numpy_writeable():
    # Create data
    original = np.random.rand(1024, 1024)
    mutable = zarr.array(original)

    # Initialize app
    app = create_zarr_server(mutable)

    # Open remote array and compare
    remote_store = HTTPStore(TestClient(app))
    arr = zarr.open_array(remote_store)
    arr[:50, :50] = 2

    np.testing.assert_allclose(arr[:], mutable[:])


def test_nested_array():
    # Create zarr hierarchy
    original = np.random.rand(1024, 1024)
    grp = zarr.open()
    grp.create_dataset("nested", data=original)

    # Intitilize app with nested nested array
    app = create_zarr_server(grp.get("nested"))

    # Ensure indexing works
    remote_store = HTTPStore(TestClient(app))
    arr = zarr.open_array(remote_store)
    np.testing.assert_allclose(arr[:], original)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import codecs
from setuptools import setup, find_packages


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding="utf-8").read()


requirements = []
with open("requirements.txt") as f:
    for line in f:
        stripped = line.split("#")[0].strip()
        if len(stripped) > 0:
            requirements.append(stripped)


# https://github.com/pypa/setuptools_scm
use_scm = {"write_to": "simple_zarr_server/_version.py"}

setup(
    name="simple-zarr-server",
    author="Trevor Manz",
    author_email="trevor.j.manz@gmail.com",
    license="BSD-3",
    url="https://github.com/manzt/simple-zarr-server",
    description="A simple server for sharing zarr over HTTP.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=requirements,
    use_scm_version=use_scm,
    setup_requires=["setuptools_scm"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Testing",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
    ],
    entry_points={
        "console_scripts": ["simple-zarr-server = simple_zarr_server.cli:main",],
    },
)



import numpy as np
from starlette.applications import Starlette
from starlette.responses import Response
from starlette.routing import Route
from starlette.middleware.cors import CORSMiddleware

import uvicorn
import zarr


def create_zarr_server(z):
    """Creates a Starlette app, mapping HTTP requests on top of a Zarr store.

    Parameters
    ----------
    z : zarr.Array or zarr.Group
        Part of zarr hierarchy to expose over HTTP. The `read_only` property is
        checked to determine which HTTP methods are available.

    Returns
    -------
    app : starlette.applications.Starlette
        Starlette app
    """
    # Use read_only property to determine how the underlying store should be protected.
    methods = ["GET", "HEAD", "PUT"] if not z.read_only else ["GET", "HEAD"]
    # If z isn't the root of a hierarchy, need to prepend it's path.
    path_prefix = f"{z.path}/" if z.path else ""

    async def map_request(request):
        path = path_prefix + request.path_params["path"]
        if request.method == "PUT":
            # PUT only handled if writeable
            try:
                blob = await request.body()
                z.store[path] = blob
                return Response(status_code=200)
            except:
                return Response(status_code=404)
        else:
            try:
                # Return blob if GET, otherwise it's HEAD and should return empty body
                body = z.store[path]
                if request.method == "HEAD":
                    body = None
                return Response(body, status_code=200)
            except KeyError:
                # Key not in store, return 404
                return Response(status_code=404)

    routes = [Route("/{path:path}", endpoint=map_request, methods=methods)]
    return Starlette(routes=routes)


def serve(source, *, allowed_origins=None, **kwargs):
    """Starts an HTTP server, serving a part of a zarr hierarchy or numpy array as zarr.

    Parameters
    ----------
    source : zarr.Array, zarr.Group, or np.ndarray
        Source data to serve over HTTP. The underlying store of a zarr.Array,
        or zarr.Group are used to forward requests. If a numpy array is provided,
        an in-memory zarry array is created, and the underlying store is wrapped.
    allowed_origins : list of str, optional
        List of allowed origins (as strings). Use wildcard "*" to allow all.
    **kwargs : keyword arguments
        All extra keyword arguments are forwarded to uvicorn.run
    """

    if isinstance(source, np.ndarray):
        # Need to cast as zarr and create store for in memory numpy array
        source = zarr.array(source)

    if not isinstance(source, (zarr.Array, zarr.Group)):
        raise TypeError(
            "Source is not one of numpy.ndarray, zarr.Array, or zarr.Group."
        )

    server = create_zarr_server(source)
    if allowed_origins:
        server.add_middleware(
            CORSMiddleware,
            allow_origins=allowed_origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
    uvicorn.run(server, **kwargs)
